/* N-body calculation using Greenspan 1st-order discrete mechanics.    */
/* Incorporates xyz-parallelism, uses domain pointers, combines        */
/* particle and pair domains, diagonal processors store particles      */
/* Spreading and accumulation coded in PARIS, more globals,            */
/* Booleans stored in char fields.                                     */
/* PARIS recoded to V5 standards				       */

/*    Provides for insertion of asteroids uniformly spaced in circular */
/*    orbit of specified radius, control over printing.                */

#include <stdio.hs>
#include <cm/cm.hs>
#include <math.hs>
#include "cmtime.hs"

/* Max number of bodies in the calculation.  N = 64 fits in 16k procs
   N must be a power of 2, 4N^2 <= no. of v. procs.  LLN = log2(N)    */

#define N  64
#define LN  6

int NB;      /* Number of bodies in system, <= N */
int LNB;     /* Largest power of 2 < NB */
int LLNB;    /* log2(LNB) */
float G;     /* Gravitational constant */

#define TWOPI 6.2831853

int PX;      /* Print control */
int NX;      /* Number of bodies given explicitly */
int NA;      /* Number of asteroids */
float AR;    /* Asteroid orbit radius */

/* Pair domain:								*/
/* Particles, and pairs of particles.  Diagonal entries have particle	*/
/* states.  Some variables could be overloaded. 			*/
/* act  - processor is active pair					*/
/* here - processor represents particle					*/
/* sp   - processor participates in spreading                           */
/* smk  - control for spreading, accumulation, when procs change        */
/* psym - plot symbol							*/
/* self - self-address for manipulation                                 */

/* VT-100 plotting for N-body programs - Capital letters leave trails	*/

/* Maximum screen size */

#define SXSZ 45
#define SYSZ 90

int SX=22, SY=80;	/* Actual size, default initialization */
/* SX should leave room below display for time line followed by NL */

/* The order of the next four declarations is essential to obtain correct
   cube addresses for pair.  */

domain screen { int sx, sy;  char oldc, newc, chng; };

domain pair { int c, i, j, smk;  char act, sm, here, sp, smm, psym;
	      CM_cubeaddr_t self;
	      float x0, x1, v0, v1, m, d0, r0, ac, w;
	      domain pair *r; };

domain screen Sc[SXSZ][SYSZ];

domain pair P[3][N][N];

float XL, XR, YB, YT;  /* Screen limits for physical coordinates */
int scgood;            /* Display assumed to correspond to Sc    */

/* Cm addresses for PARIS code */

CM_memaddr_t smk_addr, sp_addr, smm_addr, self_addr, ac_addr, w_addr;


/* Pairs will be represented symmetrically.  The function setup
   (below) initializes these so that pi, pj point to appropriate
   particle in corresponding coordinate.   */

void pair::setup()
 { int offset;  mono int k;
   for ( LNB = 1, k = 0;  LNB < NB;  LNB <<= 1, k++ ) ;
   sp = (i<LNB) && (j < LNB);  LNB >>= 1;  LLNB = k-1;
   self = (CM_cubeaddr_t)this;
   smk_addr = &smk;  smm_addr = &smm;  sp_addr = &sp;
   self_addr = &self;  ac_addr = &ac;  w_addr = &w;
   offset = (this - &P[0][0][0]);
   c = offset/(N*N);  i = (offset/N)%N;  j = offset%N;
   sm = (i < NB) && (j < NB);
   act = sm && (i != j);  here = sm && (i == j);
   r = (c<2) ? this+N*N : this-2*N*N;
   smm = 1;  smk = i^j;  /* xor */
   for ( k = 1; k < NB; k <<= 1 )
    if (smm && (k & smk)) { smk = k;  smm = 0; }
   if ( !sp ) smk = 0;
 }

/* Insert asteroids.  This assumes that orbit of radius 1 about
   origin corresponds to velocity 2*pi (e.g., au-earth-yr units)
   This version avoids complications with parallel trig.         */

void mkast()
 { float sx, cx, spd;  double arg;  int k;
   NA = NB - NX;
   if ( NA > 0 )
    { spd = TWOPI/sqrt(AR);
      for ( k = NX;  k < NB;  k++ ) 
       { arg = (TWOPI*(k-NX))/NA;
         sx = sin(arg);
         cx = cos(arg);
	 P[0][k][k].x0 =  cx;  P[0][k][k].v0 = -sx;
	 P[1][k][k].x0 =  sx;  P[1][k][k].v0 =  cx;
	 P[2][k][k].x0 = 0.0;  P[2][k][k].v0 = 0.0;
        }
      [domain pair].
       if ( here && (i >= NX) )
	 { x0 *= AR;  v0 *= spd;  m = 1.0/NA; }
     }
 }

FILE *datap;      /* Stream for reading data */

/* Load inital positions, velocities, also masses, gravitational const. */

void pair::init()
 { mono int k;  mono float a, b, c;
   for ( k = 0; k < NB; k++ )
    { fscanf(datap, "%e%e%e", &a, &b, &c);
      P[0][k][k].x0 = a;  P[1][k][k].x0 = b;  P[2][k][k].x0 = c;
    };
   for ( k = 0; k < NB; k++ )
    { fscanf(datap, "%e%e%e", &a, &b, &c);
      P[0][k][k].v0 = a;  P[1][k][k].v0 = b;  P[2][k][k].v0 = c;
    };
   for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e", &a);
       P[0][k][k].m = P[1][k][k].m = P[2][k][k].m = a;
     };
   fscanf(datap, "%e", &G);
 }

void pair::print_state()
 { mono int k;
   printf(
  "\n       x         y         z         vx        vy        vz        m\n");
   for ( k = 0; k < NB; k++ )
     printf("%2d %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.2e\n",
            k, P[0][k][k].x0, P[1][k][k].x0, P[2][k][k].x0,
               P[0][k][k].v0, P[1][k][k].v0, P[2][k][k].v0, P[0][k][k].m);
 }

/* smk[i,j] is a power of 2 corresponding to the stage at which proc[i,j]
   sends it's acceleration to partner and drops out of accumulation. It
   also corresponds to the stage at which proc[i,j] begins to send values
   during spreading operations.                                          */

/* Spread xv from diagonal along i-axis.  Leaves result in w. */
/* May spread to power of 2 just larger than NB.              */

void pair::spreadi(xv)
  float xv;
  { mono int k, a, b;  mono CM_memaddr_t xv_addr;
    xv_addr = &xv;  smm = here;
    a = (int)smk_addr + LLNB + 1;  b = (int)self_addr + LN + LLNB;
    for ( k = LLNB+1;  k > 0;  k--, a--, b-- )
    { CM_load_context(smm_addr);
      CM_logior_context((CM_memaddr_t)a);
      CM_u_move_constant_1L(smm_addr, 1, 1);
      CM_lognot_1_1L((CM_memaddr_t)b, 1);
      CM_f_move_1L(w_addr, xv_addr, 23, 8);
      CM_send_1L(xv_addr, self_addr, w_addr, 32, CM_no_field);
      CM_lognot_1_1L((CM_memaddr_t)b, 1);
      CM_load_context(sp_addr);
     }
    w = xv;
  }

/* Spread xv from diagonal along j-axis.  Leaves result in w.  */
/* May spread to power of 2 just larger than NB.               */

void pair::spreadj(xv)
  float xv;
  { mono int k, a, b;  mono CM_memaddr_t xv_addr;
    xv_addr = &xv;  smm = here;
    a = (int)smk_addr + LLNB + 1;  b = (int)self_addr + LLNB;
    for ( k = LLNB+1;  k > 0;  k--, a--, b-- )
    { CM_load_context(smm_addr);
      CM_logior_context((CM_memaddr_t)a);
      CM_u_move_constant_1L(smm_addr, 1, 1);
      CM_lognot_1_1L((CM_memaddr_t)b, 1);
      CM_f_move(w_addr, xv_addr, 23, 8);
      CM_send_1L(xv_addr, self_addr, w_addr, 32, CM_no_field);
      CM_lognot((CM_memaddr_t)b, (CM_memaddr_t)b, 1);
      CM_load_context(sp_addr);
     }
    w = xv;
  }

/* Sum accelerations.  Called in context sp */

void pair:: sumac()
 { mono int k, a, b;
   smm = 1;  w = 0.0;
   a = (int)self_addr;  b = (int)smk_addr;
   /* Initial context is sp */
   for ( k = 0;  k < 1 + LLNB;  k++, a++, b++ )
    { CM_logand_context((CM_memaddr_t)b);       /* Select senders */
      CM_u_move_zero_1L(smm_addr, 1);           /* Delete senders */
      CM_lognot_1_1L((CM_memaddr_t)a, 1);       /* Target address */
      CM_send_1L(w_addr, self_addr, ac_addr, 32, CM_no_field);
      CM_lognot_1_1L((CM_memaddr_t)a, 1);       /* Restore address */
      CM_load_context(smm_addr);                /* Select receivers */
      ac += w;
     }
   CM_load_context(sp_addr);
 }

/* Calculate accelerations - called with sp's active */

void pair::accel()
 {    float d1, r1;  mono int k;
      ac = 0.0;  spreadi(x1);  d1 = w;  spreadj(x1);
      if (act)
      { d1 -= w;  r1 = d1*d1;  r->ac = r1;
        ac += r1;  r->w = ac;  r1 = sqrt(w+r1);
        ac = G*m*(d1+d0)/(r0*r1*(r0+r1)); };  /* end if (act) ... */
      sumac();
 }

/* Calculate initial accelerations */

void pair::accel1()
 {    char smm;  mono int k;
      ac = 0.0;  spreadi(x0);  d0 = w;  spreadj(x0);
      if (act)
      { d0 -= w;  r0 = d0*d0;  r->ac = r0;
        ac += r0;  r->w = ac;  w += r0;  r0 = sqrt(w);
        ac = G*m*d0/(r0*w); };  /* end if (act) ... */
      sumac();
 }

/* Greenspan's 1st-order method.  Performs ns steps of time dt
   with solution tolerance eps.  Returns number of evaluations.
   Leaves number of iterations which were stopped at limit in cfail.
   Some printing for debugging is provided in comments.  */

int cfail;   /* Counts number of iterations that failed to converge */

int green1(ns, dt, eps)
  int ns;  float dt, eps;
 { int k, ic, v, lim=10;  float hdt, mx, my;
   v = 0;  hdt = 0.5*dt;
   [domain pair]. if (sp)  /* Everything in "spreading" procs */
   { float nx;
     spreadi(m);  m = w;
     for (k = 0; k < ns; k++)
       { ic = 0;  accel1();
	 /* This is a do - while - repeat loop: */
   	 do
          { v++;  ic++;
	    if (here)
	    { nx = v0 + dt*ac;  mx = (>?= fabs(nx - v1));  v1 = nx;
	      nx = x0 + hdt*(v0 + v1);
	      my = (>?= fabs(nx - x1));  if (my > mx) mx = my;
	      x1 = nx; }
	    if ( (mx < eps) || ( ic == lim ) ) break;
	    accel(); }
	 while (1);
         if (here) { x0 = x1;  v0 = v1; };  if (ic == lim) cfail++;
       } ;   /* end for */
    }  /* end  [domain pair] */
   return(v);
 }

/* Read screen limits and plot symbols from datap */

void loadsym()
 { int k;  char c[2];
   fscanf(datap, "%e%e%e%e", &XL, &XR, &YB, &YT);
   [domain pair].
    { for ( k = 0;  k < NX;  k++ )
       { fscanf(datap, "%1s", c);  P[0][k][k].psym = c[0]; };
      if ( here && (i >= NX) ) psym = 'x';
    }
 }

/* Initialize screen */

void initplot()
 { [domain screen].
    { int offset;
      offset = this - &Sc[0][0];  sx = offset/SYSZ;  sy = offset%SYSZ;
      oldc = '\x20';  newc = '\x20';
     }
 }

/* Send current state to screen and update */

void send_state()
 { float scx, scy;
   scx = (SY - 1)/(XR - XL);  scy = (SX - 1)/(YB - YT);
   [domain pair].
    { int u, v;
      if ( here )
       { u = -10;  v = -10;
	 if ( (c == 0) && (XL <= x0) && (x0 <= XR) )
	    { u = (int)((x0 - XL)*scx);
	      if (u < 0) u = 0; else if (u >= SY) u = SY - 1; };
	 if ( (c == 1) && (YB <= x0) && (x0 <= YT) )
	    { v = (int) ((x0 - YT)*scy);
	      if (v < 0) v = 0; else if (v >= SX) v = SX - 1; };
	 if ( (c == 0) && (u >= 0) )
	    { v = r->v;  if ( v >= 0 ) Sc[v][u].newc ,= psym; }
	}
    };
   [domain screen].
    { chng = 0;
      if ( (oldc != newc) && !((oldc == '.') && (newc == '\x20')) )
        { if ( ('A' <= oldc) && (oldc <= 'Z') && (newc == '\x20') )
	     oldc = '.'; else  oldc = newc;
	  chng = 1; };
      newc = '\x20';
    }
 }

/* Refresh display, ignoring previous condition.
   'tm' is the current simulation time.	    	 */

void refresh(tm)
 float tm;
 { int u, v;
   if (PX >= 4) printf("\x1B[1;1H"); else printf("\n");
   [domain screen].
    { for ( u = 0;  u < SX;  u++)
       { for ( v = 0;  v < SY;  v++ ) printf("%c", Sc[u][v].oldc);
	 printf("\n"); };
      chng = 0;
     }            /* Following has an innocuous newline */
   printf("Time:  %9.2f\x1B[1A\n\x1B[20C", tm);
 }     						

/* Update display, assuming display reflects previous screen.
   'tm' is the current simulation time.  Uses innocuous newlines
   to force output.						 */

void update(tm)
 float tm;
 { int u, v, k = 0;
   [domain screen].
    while ( chng )
      { u = (<?= sx);  if ( sx == u ) v = (<?= sy);
	printf("\x1B[%d;%dH%c", u+1, v+1, Sc[u][v].oldc);
	Sc[u][v].chng = 0;
	k++;  if ( k == 10 ) { printf("\n");  k = 0; }; };
   printf("\x1B[%d;9H%9.2f\x1B[1A\n\x1B[20C", SX+1, tm);
 }

/* Print current state under control of PX.  Adjusts NB to control
   pair::print_state.                                               */

void ax_pstate(tm)
 float tm;
 { mono int k, np;
   np = (PX == 0) ? 0 : (PX == 1) ? NX : NB;
   if ( (PX < 3) && (np > 0) )
    { k = NB;  NB = np;  [domain pair].{ print_state();};  NB = k; }
   else if ( PX >= 3 )
    { send_state();  /* Use refresh for now */
      if ( PX <= 4 || !scgood ) refresh(tm); else update(tm);
      scgood = 1; }
 }

/* Commands are:
   n <n>     - set total number of bodies
   p <n>     - set print code to N:
	       0 - nothing, 1 - NX bodies, 2 - all,
	       3 - plot with rewrite, 4 - plot with overwrite
               5 - plot with random access
   r         - run calculation, prompts for data file
   s <h> <w> - set screen size to h by w (max 45 x 90)
   q         - quit                                    */

void main()
 { int ns, np, cmd, nv, k, vl;  float dt, eps;
   char fnm[50], c[2];
   PX = 2;
   while (printf("nbax> "),  scanf("%1s", c),  cmd = c[0],  cmd != 'q' )
    if ( cmd == 'n' ) scanf("%d", &NB); else
    if ( cmd == 'p' ) scanf("%d", &PX); else
    if ( cmd == 's' ) scanf("%d%d", &SX, &SY); else
  { /* Performs np integrations, each ns steps of dt */
     printf("Data file> ");  scanf("%s", fnm);
     datap = fopen(fnm, "r");
     fscanf(datap, "%d%e", &NX, &AR);
     [domain pair].
       { setup();  nv = NB;  NB = NX;  init();  NB = nv; };
     mkast();
     fscanf(datap, "%d%d%e%e", &ns, &np, &dt, &eps);
     if ( PX >= 3 ) { loadsym();  initplot();  scgood = 0; };
     fclose(datap);
     printf("\n   G = %13.6e  delta-t = %12.5e  epsilon = %12.5e\n",
                   G, dt, eps);
     printf("   %d asteroids at radius = %9.5f\n", NA, AR);
     printf("\n time = 0");  ax_pstate(0.0);
     for ( k = 0; k < np; k++ )
      { if ( PX < 3) CM_start_timer(1);
        cfail = 0;
        vl = green1(ns, dt, eps);
        if ( PX < 3)
	 { CM_stop_timer(1);  CM_reset_timer();
           printf(
	 "\n simulated time = %9.3f  %d evaluations  %d iterations limited\n",
	  (k + 1)*ns*dt, vl, cfail);  ax_pstate(0.0); }
        else ax_pstate((k + 1)*ns*dt);
      }
   }
 }

/* Arrangement of data file:

    NX			 Number of bodies specified explicitly in file 
    AR			 Radius of asteroid orbits

     x      y      z
     .      .      .     Initial positions
     .      .      .     for NX points
     .      .      .

    vx     vy     vz
     .      .      .     Initial velocities
     .      .      .
     .      .      .

     m
     .                   Masses
     .
     .

     G                   Gravitational constant

    ns   np   dt   eps   Integration parameters

 { Remaining items read only if plotting selected: }

    XL   XY   YB   YT    Plotting limits (simulation coordinates)

    EMJSUN...            Plot symbols                            */
